# xrxs

`xrxs` is an experimental game server using the Plan 9 protocol `9p`.

The client is intended to be a specialized [uxn](https://wiki.xxiivv.com/site/uxn.html) ROM that can load other ROMs (possibly preserving its own code to maintain a common menu system or interface to `xrxs`).

## design

This is the working structure of the 9p filesystem:

* `/ctl`: Write-only control file for inputing system commands (should we check the /realms file for success or failure, or should we make ctl r/w and return success or failure of the current user's last command on read?)
  * `login PW`: Authenticate with `xrxs` -- password is hashed against realm password hash
  * `load CART`: Load a cartridge
  * `create REALM`: Create a new realm (start a new game)
  * `protect REALM PW`: Protect the realm with a password
  * `enter REALM`: Join an existing realm
  * `leave`: Leave the current realm
  * `save`: Save the current universe to the realm (redundant? or force universe/realm to flush?)
  * `reset`: Reset the cartridge (attempts to leave the realm, but doesn't save)
  * `unload`: Unload the cartridge
  
* `/users`: Read-only; Self and others in the realm are readable from here, one per line. It contains only yourself before joining a realm. Your username on your machine is used as your username in `xrxs` -- if your name is taken, you will get an error on attaching.

* `/carts`: Available game/app cartridges for this server, read only; Carts are listed per line upon reading the file. It is backed by files on the server in a directory structure like `carts/<cart_name>/{<cart_name.rom>, data/}`.
  
* `/slot`: After loading the cartridge, its ROM is read from here; Read-only.

* `/data/`: Any supporting data that comes with the cartridge will be found here; read only

* `/realms`: Open/saved realms, read-only. Realms and their associated universe are backed by real files on the server so that they can be preserved across service instantiations, in a directory structure like: `realms/<realm_name>/{realm, universe}`. Realms can either be solo, open, or protected; Open or protected realms can have limited member numbers. Depending on the cartridge, these settings can be user-managed or managed by the cartridge itself. Realms are listed per line upon reading the file like: `realmx 1 4 1`. `realmx` would be the name of the realm. The first number is number of members, second is member limit, third is 1 if protected, 0 if not. `0 1 1` represents a protected solo realm that is empty (saved game with password). `0 1 0` represents an unprotected solo realm that is empty (saved game with no password).

* `/universe`: Write here to update serverside state for this cart/realm; Read from here to get the current state. Each cartridge will have its own format for encoding data here. Since this changes frequently in multiplayer games, perhaps only flush to disk at regular intervals instead of after every write.
  
## realm format

Each realm directory on the server should have the following files:

  * `realm`: Basic data for the realm, file should contain only the maximum number of members, and the password hash, if any, separated by a space.
  * `members`: Members currently in the realm, one per line.
  * `universe`: The actual game state in for the realm, format is cartridge-specific.
  
Everything in the realm directory is synchronized to disc both when realm membership, limit, or password changes, and at regular intervals. If a member makes no writes to the universe after a certain time, consider them gone from the realm (so there should be a common SYN message all cartridges implement to keep them alive in the realm).